# -*- coding: utf-8 -*-
"""
Created on Tue Oct 16 14:33:20 2018
@author: js

Inspired by Micasense RedEdge alignment algorithms.

"""

#%% Imports

import numpy as np
import tkinter  as tk
from tkinter import filedialog
import os
import spectral as sp
import cv2

#%% Functions for file/folder selection 
'''
# Ask the user to select directory.
'''
def select_directory(text_title):
    
    root = tk.Tk()
#    root.withdraw() #use to hide tkinter windo  
    path = filedialog.askdirectory(title= text_title)
    root.destroy()
    return path 

# '''
# # Ask the user to select a single file.
# '''
# def select_file(text_title):
    
#   root = tk.Tk()
#   root.title(text_title)
# #  root.withdraw() #use to hide tkinter windo 
  
#   fname = filedialog.askopenfilename (title= text_title,
#                            filetypes=(("All files", "*.*"),
#                                    ("TXT files",".txt") ))
#   root.destroy()
#   return fname

'''
# Ask the user to select files.
'''
def select_files(text_title):
    
  root = tk.Tk()
  root.title(text_title)
#  root.withdraw() #use to hide tkinter windo 
  
  fname = filedialog.askopenfilenames(title= text_title,
                           filetypes=(("All files", "*.*"),
                                   ("TXT files",".txt") ))
  root.destroy()
  return fname

#%% Function to detect edges on the datacube
    
def edge_image(DataCube, WindowSize=3):
    # init output datacube
    EdgeCube = np.zeros(DataCube.shape, dtype=np.uint8)
    # process each spectral band separately
    for i in range(DataCube.shape[2]):
        # blur image for better edge detection
        ImgBlur = cv2.GaussianBlur(DataCube[:,:,i], (WindowSize,WindowSize), 0)
        # use OpenCV Sobel edge detection algorithm
        EdgeImg_x = cv2.Sobel(src=ImgBlur, ddepth=cv2.CV_64F, dx=1, dy=0, ksize=5)
        EdgeImg_y = cv2.Sobel(src=ImgBlur, ddepth=cv2.CV_64F, dx=0, dy=1, ksize=5)
        EdgeImg = cv2.addWeighted(np.abs(EdgeImg_x), 0.5, np.abs(EdgeImg_y), 0.5, 0)
        
        # Scale to 8-bit image
        EdgeImg *= 255 / np.max(EdgeImg)
        # save to output datacube
        EdgeCube[:,:,i] = EdgeImg.astype(np.uint8)
    
    return EdgeCube

#%% Function to detect the relative band image alignments and to produce the aligned datacube 
    
def align_datacube(DataCube, RefBand=1):
    # define warp model
    WarpMode = cv2.MOTION_EUCLIDEAN
    WarpMatrix0 = np.array([[1,0,0], [0,1,0]], dtype=np.float32)
    # Terminate the optimizer if either the max iterations or the threshold are reached
    Epsilon=1e-9
    MaxIterations=40
    TermCriteria = (cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT, MaxIterations, Epsilon)
    
    # detect edges in the image. This makes the different bands more equivalent
    EdgeCube = edge_image(DataCube, WindowSize=11)
    
    # Init aligned datacube
    AlignedDataCube = DataCube.copy()
    
    # process each spectral band
    for BandNum in range(DataCube.shape[2]):
        print('Aligning band {:.0f}.'.format(BandNum))
        if BandNum != RefBand:
            # Find warp parameters
            _, WarpMatrix1 = cv2.findTransformECC(EdgeCube[:,:,RefBand], 
                                                  EdgeCube[:,:,BandNum], 
                                                  WarpMatrix0, 
                                                  WarpMode, 
                                                  TermCriteria, 
                                                  inputMask=None, 
                                                  gaussFiltSize=1)
            # Warp the image
            AlignedDataCube[:,:,BandNum] = cv2.warpAffine(DataCube[:,:,BandNum],
                                                          WarpMatrix1,
                                                          DataCube.shape[:2],
                                                          flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP)
    return AlignedDataCube


#%% Define input files and output folder

# Ask user to select raw datacube input files
HDRFiles = select_files('Select HDR file(s)')

# Ask user to select output folder
OutputFolder = select_directory('Select output folder')
print('\n->Output directory: ', OutputFolder)
print('\n')


#%% Process each selected file

for HDRFile in HDRFiles:
    # Show progress
    print('Processing: ' + os.path.split(HDRFile)[-1])

    # load in the unaligned datacube
    DataCube = sp.envi.open(HDRFile).read_bands(np.arange(46))
    
    # Align it
    AlignedDataCube = align_datacube(DataCube, RefBand=24)

    # Form output file name
    OutputHDRFile = os.path.join(OutputFolder, os.path.split(HDRFile)[1].replace('.hdr', '_aligned.hdr'))
    
    # Save aligned datacube as ENVI BSQ
    sp.envi.save_image(OutputHDRFile, AlignedDataCube, dtype=np.float32, ext='bsq', interleave='bsq', force=True)



